#
# Auto select of BET range using Rouquerol criteria:
#
# C > 0
# n(min P) < n(monolayer) < n(max P) & min(P) < P(nm) < max(P)
# n*(1-p/po) is increasing from n(min P) to n(max P)
# R2 > 0.995
# p/po(exp) ~ 1/(sqrt(C) + 1)
#
# Micromeritics Instrument Corp.
# July 2013
#
# Jeff Kenvin
#
# This script my be distributed with the inclusion of this header.
#
import mic
import numpy as np
#import matplotlib.pyplot as plt
#
v2m = 1.0/22414 # mol/cm3
R2min = 0.9975
#
prel, qads, n_ads, warm_fs, cold_fs, mass, desc = mic.isotherm('rel')
#
csa, hsd, dcf, mw, ag = mic.adsorptive_data()
#
p = prel[0:n_ads]
v = qads[0:n_ads]
#
n = v*v2m
pMin = 0.0001
pMax = 0.305
#
# function for specific surface area calculation
#
def fSSA( n_mono,csa ):
    An = 6.02214129e23 # mol-1
    a = An*n_mono*csa*1.0e-18
    return a
#
# Rouquerol transform
#
def fRouq( x,n ):
    rq = n*(1.0 - x) #Rouquerol transform
    rqMax = np.max(rq)
    iPmax = np.nonzero( rq == rqMax )
    iR = np.nonzero( (rq[0:-1] < rq[1:]) & (x[:-1] <= x[iPmax]) )# monotonic and increasing
    maxIR = np.max(iR) + 1 
    iR = np.append(iR , maxIR)
    pRQ = x[iR]
    nRQ = n[iR]
    rqG = nRQ*(1.0-pRQ) # best Rouquerol values
    nRQm = np.min(nRQ) # minimum
    nRQM = np.max(nRQ) # maximum
    return rq,pRQ,nRQ,nRQm,nRQM,rqG
#
# BET transform
#
def fBET( x,n ):
    yBET = 1.0/n/(1.0/x - 1.0)
    return yBET
#
# Calculate BET parameters
#
def fCalcBET( x, yBET ):
    ix = np.ones(x.shape)
    A = np.array([x,ix])
    mb = np.linalg.lstsq(A.T,yBET)[0] # y = Ax
    nM = 1.0/(mb[0] + mb[1]) # monolayer capacity
    C = 1.0 + mb[0]/mb[1] # BET C constant
    corcoef = np.corrcoef(x,yBET)[0,1]
    rSQ = corcoef**2
    return nM,C,rSQ,mb[0],mb[1]
#
# Select a subset of data
#
def fSelect(x,n,pMin,pMax):
    iSelect = np.nonzero( (x >= pMin) & ( x<= pMax ) )
    pS = x[iSelect]
    nS = n[iSelect]
    return pS,nS
#
# Calculate relative pressure at monolayer
#
def fPcalc(n,slope,intercept): # x/(n(1-x)) = mx + b
    a = slope
    b = 1.0/n - slope + intercept 
    c = -intercept
    x1 = -b + (b*b - 4.0*a*c)**0.5
    x1 = x1/2.0/a
    x2 = -b - (b*b - 4.0*a*c)**0.5
    x2 = x2/2.0/a
    return x1,x2
#
# Calculate n_ads from BET constants
#
def fNcalc(x,m,b): # calculate n for slope and intercept
    invN = m - m*x + b/x - b
    nOut = 1.0/invN
    return nOut
#
def fBestBET(p,n,pMin,pMax): # calculate best BET and return params for optimization
    # pMin and pMax provide a wide starting range for selecting BET points
    pR,nR = fSelect(p,n,pMin,pMax)
    # Use the Rouquerol transform for selecting monotonic points
    rq,pRQ,nRQ,nRQm,nRQM,rqG = fRouq(pR,nR)
    # calculate BET transform for Rouquerol points
    yBET = fBET(pRQ,nRQ)
    # calculate BET parameters
    nM,C,rSQ,slope,intercept = fCalcBET( pRQ, yBET )
    # calculate pressure for nMono
    pM1,pM2 = fPcalc(nM,slope,intercept)
    return pR,nR, rq, pRQ, nRQ, nRQm, nRQM, rqG , yBET, nM, C, rSQ, slope, intercept,pM1
#
pR0,nR, rq0, pRQ, nRQ, nRQm,nRQM, rqG , yBET, nM, C, rSQ, slope, intercept,pM1 = fBestBET(p,n,pMin,pMax)
pminOpt = np.max(pRQ)/11.0
pR,nR, rq, pRQ, nRQ, nRQm, nRQM, rqG , yBET, nM, C, rSQ, slope, intercept,pM1 = fBestBET(p,n,pminOpt,pMax)
#
#
# check for valid n-monolayer, Rsq, C
isNok = (nM>=nRQm) & (nM<=nRQM)
isR2ok = (rSQ >= R2min)
isCok = ( C > 0.0 )
isOK = isCok & isR2ok & isNok & True
# calculate surface area
aBET = fSSA(nM,csa)
#
# plot BET
#
mic.graph("BET transform","p/po","x/(n*(1-x))")
mic.graph.add("Selected",pRQ,yBET,marker='o',graphtype='points')
xLine = np.array([0.0,np.max(pRQ)])
yLine = slope*xLine + intercept
red=[255,0,0]
mic.graph.add("Fit",xLine,yLine,linestyle='-',graphtype='curve',color=red)
yNm = slope*pM1+intercept
blk=[0,0,0]
mic.graph.add("Nm",pM1,yNm,marker='x',graphtype='points',color=blk)
#
RQc = 1.0/(C**0.5 + 1.0)
isPok = ((RQc-pM1)**2)**0.5/pM1*100.0 < 5.0
pRQm = np.min(pRQ)
pRQM = np.max(pRQ)
#
mic.summary(title='Summary')
nMmm = nM*1000.0
nRQmmm = nRQm*1000.0
nRQMmm = nRQM*1000.0
#
mic.summary.add('BET Optimized Summary',['Surface Area, m2/g', 'C' ,'Rsq', '-----',
                                'P/Po(minimum)', 'P/Po(@ monolayer)',
                                '1/('+'\u221A' + 'C'+'\u0305' + '+1) ', 
                                'P/Po(maximum)',
                                'n(min P/Po), mmol/g', 'monolayer, mmol/g',
                                'n(max P/Po) mmol/g',"-----",
                                'slope g/mol','intercept g/mol',
                                "-----", 'C > 0','n(min) < monolayer < n(max)',
                               'P/Po(mono) ~1/('+'\u221A' + 'C'+'\u0305' + '+1) ' ,
                                'Linear '],
                [ "%8.2f" % aBET,"%6.2f" % C, "%6.5f" % rSQ, '-----' ,
                  "%5.6f" % pRQm,
                  "%5.6f" % pM1, "%5.6f" % RQc,
                  "%5.6f" % pRQM,
                 "%5.6f" % nRQmmm, "%5.6f" % nMmm, "%5.6f" % nRQMmm,
                  "-----",
                 "%5.3f" % slope, "%5.4f" % intercept, "-----",
                  isCok,isNok,isPok,isOK])
#
nOut = fNcalc(pR,slope,intercept)
#
# plot Rouquerol
#
mic.graph("Rouquerol transform","p/po","n*(1-x)",xlinear=False)
mic.graph.add("Data",pR0,rq0,marker='o',graphtype='points')
mic.graph.add("Selected",pRQ,rqG,marker='x',graphtype='points')
#
